use v6.d;

unit module Cryptoish;

constant @base40  := (0 .. 9, "A" .. "Z", " ", ".", "?", "!").flat;
constant $base    := 400;
constant $wrapper := 36;

our subset Key      of Str where { /^ @base40 ** 2 $/  };  ## NEW
our subset Alphabet of Str where { /^@base40+$/  };
our subset Modulo   of Int where { 2 <= $_ <= 40 };
our subset Shuffle  of Int where { $_ == 0 || 2 <= $_ <= 40 };

sub load-it ($secret = "")
{
  my $name = $secret ?? "Cryptoish::Secret::$secret" !! "Cryptoish::Secret";
  require ::($name) '&get-secret';
  return &get-secret;
}

our sub get-alphabet
{
  return @base40;
}

our sub encrypt (Alphabet $key, Alphabet $text, :$secret)
{
  my &get-secret = load-it($secret);

  my $idx = 0;
  my %values;
  %values{$_}.push($++) for &get-secret($key).comb;

  return $text.flip.comb.map( { %(%values).{$_}.pick * $base ** $++ } ).sum.base($wrapper);
}

our sub decrypt (Alphabet $key, Str $value, :$secret)
{
  my &get-secret = load-it($secret);
  
  my $result = "";

  my $number = $value.parse-base($wrapper);
  while $number
  {
    $result ~= @( &get-secret($key) ).substr($number % $base, 1);
    $number = $number div $base; 
  }
  return $result.flip;
}

our sub multiencrypt (Alphabet $key is copy, Alphabet $text is copy, Modulo :$modulo = 40,
                      Bool :$flip = False, :$secret, :$verbose = False, Shuffle :$shuffle = 0)
{
  my &get-secret = load-it($secret);
  
  sub random_key
  {
    return (0 .. 9, "A" .. "Z").flat.roll(2).join;
  }

  my @chars = $text.comb;

  my @return;
  
  while @chars
  {
    my $idx = 0;
    my %values;
    %values{$_}.push($idx++) for &get-secret($key).comb;

    my $count = (0 .. 39).pick;

    print "Count: $count" if $verbose;

    @return.push(%values{@base40[$count]}.pick);

    print " ('", @base40[$count], "' -> ", @return[*-1], ")" if $verbose;
    
    $count = $count % $modulo;

    if $verbose
    {
      print " (modulo: $count)" if $modulo != 40;
      print "\n";
    }

    if $count
    {
      my @partial = @chars.splice(0, $count);

      print "Partial: '{ @partial[].join }' -> " if $verbose;

      @return.push(%values{$_}.pick) for @partial;

      say @return[*-@partial .. *-1] if $verbose;
    }
    
    if @chars
    {
      my $new-key = random_key;

      print "Key: $new-key" if $verbose;

      @return.push(%values{$new-key.substr(0,1)}.pick);
      @return.push(%values{$new-key.substr(1,1)}.pick);

      say " -> ({ @return[*-2]} { @return[*-1] })" if $verbose;

      $key = $new-key;
    }
  }

  if $verbose
  {
    say "All: ({ @return })";

    say "All (flip): ({ @return.flip })" if $flip;

    if $flip
    {
      say "Int: ", @return.reverse.map( { $_ * $base ** $++ } ).sum;
      say "Str: ", @return.reverse.map( { $_ * $base ** $++ } ).sum.base($wrapper);

    }
    else
    {
      say "Int: ", @return.map( { $_ * $base ** $++ } ).sum;
      say "Str: ", @return.map( { $_ * $base ** $++ } ).sum.base($wrapper);
    }
  }

  my $return =  $flip
    ?? @return.reverse.map( { $_ * $base ** $++ } ).sum.base($wrapper)
    !! @return.map( { $_ * $base ** $++ } ).sum.base($wrapper);

  return shuffle-string($return, $shuffle) if $shuffle;
  return $return;  
}

my %base40values = @base40.map( { $_ => $++ } );


sub shuffle-string ($string, $count)
{
  my $new = "";
  loop
  {
    state $pos = 0;
    my $partial = $string.substr($pos, $count) // last;
    $new ~= $partial.flip;
    $pos += $count;
  }
  return $new;
}

our sub multidecrypt (Alphabet $key is copy, Str $value is copy, Modulo :$modulo = 40,
                      Bool :$flip = False, :$secret, :$verbose = False, Shuffle :$shuffle = 0)
{
  my &get-secret = load-it($secret);
  
  sub char-from-key($key, $index)
  {
    return @( &get-secret($key) ).substr($index, 1);
  }

  say "Key: $key" if $verbose;
  say "Str: $value" if $verbose;

  $value = shuffle-string($value, $shuffle) if $shuffle;

  my $number = $value.parse-base($wrapper);

  say "Int: $number" if $verbose;

  my @numbers;
  while $number
  {
    @numbers.push($number % $base);
    $number = $number div $base; 
  }

  if $verbose
  {
    say "All: @numbers[]";
    say "All (flip): { @numbers.flip }" if $flip;
  }

  @numbers = @numbers.reverse if $flip;
 
  my $result = "";
  
  while @numbers
  {
    print "Count: { @numbers[0] } -> { char-from-key($key, @numbers[0])} -> " if $verbose;

    my $count = %base40values{ char-from-key($key, @numbers.shift) };

    if $verbose
    {
      print $count;
      print " (modulo: { $count % $modulo })" if  $modulo != 40;
      print "\n";
    }

    if $count
    {
      $count = $count % $modulo;

      my @partial = @numbers.splice(0, $count);

      print "Partial: ({ @partial }) -> '" if $verbose;

      $result ~= char-from-key($key, $_) for @partial;

      if $verbose
      {
        print char-from-key($key, $_) for @partial;
        print "'\n";
      }
    }
    
    if @numbers.elems > 1
    {
      my ($a, $b) = @numbers.splice(0, 2);

      my $new-key = char-from-key($key, $a) ~ char-from-key($key, $b);

      say "Key: ({ $a } { $b }) -> $new-key" if $verbose;

      $key = $new-key;
    }
  } 
  say "Text: '{ $result }'" if $verbose;
 
  return $result;
}
